from fluxt.app import Fluxt
import fluxt.operations as operations

# create a streaming app
fluxt = Fluxt(name='Word Count')

@operations.flat_map()
def tokenizer(event):
    return event.lower().split(' ')

@operations.key_by()
def key_by(event):
    return (event, 1)

@operations.reducer()
def count(key, accum, event):
    if accum is None:
        return 1

    return accum + event


@fluxt.stream()
def stream_processor(datastream):
    events = ['event', 'event text', 'event test test',
              'test text event hello hello']
    print(f'input events = {events}')

    datastream.source_from_collection(events)

    datastream.pipeline(tokenizer, key_by, count)

    datastream.print()



from fluxt import Fluxt
from fluxt.storage import Memory

import fluxt.operations as operations

storage = Memory()

fluxt = Fluxt(name='Page Views', storage=storage)

class FormatEvents(operations.MapFunction):
    def map(self, event):
        return event['page'], event['page_views']

class CountViews(operations.ReducerFunction):
    def reduce(self, key, reduced, event):
        if not reduced:
            return event

        return reduced + event

@operations.key_by()
def key(event):
    return event[0], event[1]

@fluxt.stream()
def views_per_page(ds):
    ds.source_from_collection([
      {'username': 'sam', 'page_views': 11, 'page': 'about'},
      {'username': 'sam', 'page_views': 7, 'page': 'home'},
      {'username': 'bill', 'page_views': 9, 'page': 'about'},
      {'username': 'reg', 'page_views': 5, 'page': 'home'},
    ])

    ds.print()

    ds.map(FormatEvents()) \
      .key_by(key) \
      .reduce(CountViews())


import sqlite3
import logging

from fluxt.storage.base import Base

logger = logging.getLogger(__name__)


class LocalDatabase(Base):
    def __init__(self, path, namespace):
        """ init memory storage """
        self.con = sqlite3.connect(path, isolation_level=None)
        self.namespace = namespace

        cur = self.con.cursor()

        try:
            cur.execute('create table keystore(namespace, key, value)')
        except sqlite3.OperationalError:
            logger.warning('keystore table already exists')

        cur.close()

    def reset(self):
        """ reset memory """
        cur = self.con.cursor()

        cur.execute('delete from keystore '
                    'where namespace = ?',
                    (self.namespace, ))

        cur.close()

    def set_key(self, key, value):
        """ set key value """
        cur = self.con.cursor()

        try:
            self.get_key(key)
        except KeyError:
            cur.execute('insert into keystore (namespace, key, value) '
                        'values (?, ?, ?)', (self.namespace, key, value))

            cur.close()
            return

        cur.execute('update keystore '
                    'set value = ? '
                    'where namespace = ? and key = ?',
                    (value, self.namespace, key))
        cur.close()

    def get_key(self, key):
        """ get key value """
        cur = self.con.cursor()

        res = cur.execute('select namespace, key, value '
                          'from keystore '
                          'where key = ?', (key,))

        row = res.fetchone()

        cur.close()

        if not row:
            raise KeyError

        return row[2]



from abc import abstractmethod

from fluxt.operations.base import Operation


def map_function_generator(func):
    class TempMap(MapFunction):
        def map(self, event):
            return func(event)

    return TempMap()


class MapFunction(Operation):
    """ map function class """

    def process_batch(self, events):
        """ passes a batch of events to map

        Args:
            events (list): list of events

        Returns:
            batch (list): list of modified events
        """
        return [self.map(event) for event in events]

    @property
    def type(self):
        """ return the operation type """
        return self.__class__.__base__.__name__

    @abstractmethod
    def map(self, event):
        """ implements a map function

        Args:
            event (object): event object

        Returns:
            object: event object
        """
        return event



from abc import ABC, abstractmethod


class Operation(ABC):
    """ base operation """

    def __init__(self):
        """ base operation init """
        self.storage_backend = None

    def __call__(self, event_collection):
        """ run process batch execution

        Args:
            event_collection (list): list of events

        Returns:
            event_collection: list of events
        """
        results = self.process_batch(event_collection.events)

        event_collection.events = results

        return event_collection

    @abstractmethod
    def process_batch(self, events):
        """ process the incoming event batch """
        pass

    @property
    def storage(self):
        """ operation storage class """
        return self.storage_backend

    @storage.setter
    def storage(self, storage):
        """ set operation storage class """
        self.storage_backend = storage



import pytest

from fluxt.operations import KeyByFunction, key_by
from fluxt.app.events import EventCollection


class KeyBy(KeyByFunction):
    def key_by(self, event):
        return event[0], event[1]


class BadKeyBy(KeyByFunction):
    pass


def test_key_by_call():
    key_by_function = KeyBy()
    event_collection = EventCollection(None)
    event_collection.events = [('test', 1), ('test', 1), ('hello', 1)]
    event_collection = key_by_function(event_collection)

    assert len(event_collection.events) == 3


def test_key_by_type():
    key_by_function = KeyBy()
    assert key_by_function.type == 'KeyByFunction'


def test_decorated_key_by():
    @key_by()
    def my_key_by(event):
        return event, 1

    event_collection = EventCollection(None)
    event_collection.events = ['test', 'test', 'world']

    assert my_key_by(event_collection).events == [
        ('test', 1), ('test', 1), ('world', 1)]


def test_key_by_function():
    key_by_function = KeyBy()
    event_collection = key_by_function.key_by(('test', 1))

    # will return tuple so len == 2
    assert len(event_collection) == 2
    assert event_collection == ('test', 1)


def test_key_by_function_abstract():
    with pytest.raises(TypeError):
        BadKeyBy()



import fluxt.operations as operations

from fluxt.operations.base import Operation


class DataStreamOperations:
    def pipeline(self, *args):
        """ create an operation pipeline """
        for operation in args:
            if not isinstance(operation, Operation):
                raise TypeError(f'{operation} is '
                                f'not type {Operation.__name__}')

            self.transformations.append(operation)

        return self

    def key_by(self, key_by_function):
        """ add key by

        Args:
            key_by_function (KeyByFunction): key by function

        Raises:
            TypeError: if functions is not type key by

        Returns:
            datastream (DataStream): datastream self
        """
        if not isinstance(key_by_function, operations.KeyByFunction):
            raise TypeError(f'{key_by_function} is '
                            f'not type {operations.KeyByFunction.__name__}')

        self.transformations.append(key_by_function)

        return self

    def map(self, map_function):
        """ add mapper

        Args:
            map_function (MapFunction): mapping function class

        Returns:
            datastream (DataStream): datastream self
        """

        if not isinstance(map_function, operations.MapFunction):
            raise TypeError(f'{map_function} is '
                            f'not type {operations.MapFunction.__name__}')

        self.transformations.append(map_function)

        return self

    def filter(self, filter_function):
        """ add filter function

        Args:
            filter_function (FilterFunction): filter function

        Returns:
            datastream (DataStream): datastream self
        """
        if not isinstance(filter_function, operations.FilterFunction):
            raise TypeError(f'{filter_function} is '
                            f'not type {operations.FilterFunction.__name__}')

        self.transformations.append(filter_function)

        return self

    def flat_map(self, flat_map_function):
        """ add flat map function

        Args:
            flat_map_function (FlatMapFunction): flat map function

        Raises:
            TypeError: if function is not a flat map

        Returns:
            datastream (DataStream): datastream self
        """
        if not isinstance(flat_map_function, operations.FlatMapFunction):
            raise TypeError(f'{flat_map_function} is '
                            f'not type {operations.FlatMapFunction.__name__}')

        self.transformations.append(flat_map_function)

        return self

    def reduce(self, reduce_function):
        """ add reducer functions

        Args:
            reduce_function (ReducerFunction): reducer function

        Raises:
            TypeError: if function is not reducer

        Returns:
            datastream (DataStream): datastream self
        """
        if not isinstance(reduce_function, operations.ReducerFunction):
            raise TypeError(f'{reduce_function} is '
                            f'not type {operations.ReducerFunction.__name__}')

        self.transformations.append(reduce_function)

        return self



import pytest

from fluxt.app.datastream import DataStream, DataStreamException
from fluxt.sources import CollectionSource
from fluxt.sinks import StdoutSink

import fluxt.operations as operations


class Map(operations.MapFunction):
    """ example map function """

    def map(self, event):
        return super().map(event)


class Filter(operations.FilterFunction):
    """ example filter function """

    def filter(self, event):
        return super().filter(event)


class FlatMap(operations.FlatMapFunction):
    """ example flat map function """

    def flat_map(self, event):
        return super().flat_map(event)


class Reduce(operations.ReducerFunction):
    """ example reducer function """

    def reduce(self, key, reduced, event):
        if not reduced:
            return event
        return event + reduced


class KeyBy(operations.KeyByFunction):
    """ example key by function """

    def key_by(self, event):
        return super().key_by(event)


def test_add_source():
    datastream = DataStream()
    datastream.add_source('source')
    assert datastream.source == 'source'


def test_add_sink():
    datastream = DataStream()
    datastream.add_sink('sink')
    assert datastream.sink == 'sink'


def test_source_from_collection():
    datastream = DataStream()
    datastream.source_from_collection([1, 2, 3])
    assert isinstance(datastream.source, CollectionSource)


def test_print():
    datastream = DataStream()
    datastream.print()
    assert isinstance(datastream.sink, StdoutSink)


def test_map():
    datastream = DataStream()
    datastream.map(Map())
    assert len(datastream.transformations) == 1

    with pytest.raises(TypeError) as e:
        datastream.map(Filter())

    assert "is not type MapFunction" in str(e)


def test_filter():
    datastream = DataStream()
    datastream.filter(Filter())
    assert len(datastream.transformations) == 1

    with pytest.raises(TypeError) as e:
        datastream.filter(Map())

    assert "is not type FilterFunction" in str(e)


def test_flat_map():
    datastream = DataStream()
    datastream.flat_map(FlatMap())
    assert len(datastream.transformations) == 1

    with pytest.raises(TypeError) as e:
        datastream.flat_map(Map())

    assert "is not type FlatMapFunction" in str(e)


def test_reduce():
    datastream = DataStream()
    datastream.reduce(Reduce())
    assert len(datastream.transformations) == 1

    with pytest.raises(TypeError) as e:
        datastream.reduce(FlatMap())

    assert "is not type ReducerFunction" in str(e)


def test_key_by():
    datastream = DataStream()
    datastream.key_by(KeyBy())
    assert len(datastream.transformations) == 1

    with pytest.raises(TypeError) as e:
        datastream.key_by(FlatMap())

    assert "is not type KeyByFunction" in str(e)


def test_pipeline():
    datastream = DataStream()
    datastream.pipeline(Map(), Filter(), Reduce())

    assert len(datastream.transformations) == 3

    with pytest.raises(TypeError) as e:
        datastream.pipeline(None, 'test')

    assert "is not type Operation" in str(e)


def test_execute_no_source():
    ds = DataStream()

    with pytest.raises(DataStreamException) as e:
        ds.execute()

    assert 'source not defined' in str(e)


def test_execute_no_sink():
    ds = DataStream()

    ds.source_from_collection([1, 2, 3])

    with pytest.raises(DataStreamException) as e:
        ds.execute()

    assert 'sink not defined' in str(e)


def test_execute(monkeypatch):
    agg = []

    def mock_pipe_capture(event_collection):
        for event in event_collection.events:
            agg.append(event)

    ds = DataStream()

    ds.source_from_collection([1, 2, 3])
    ds.print()

    monkeypatch.setattr(ds.sink, 'pipe', mock_pipe_capture)

    ds.filter(Filter()) \
        .map(Map())

    ds.execute()

    assert agg == [1, 2, 3]



import sys

from fluxt.sinks.base import Sink


class StdoutSink(Sink):
    def pipe(self, event_collection):
        """ send event to stdout """
        for event in event_collection.events:
            sys.stdout.write('%s\n' % str(event))



class GraphException(Exception):
    pass


def graph_generator(transformations, storage):
    """ generate a stream execution graph

    Args:
        transformations (list): list of transformations

    Returns:
        graph (StreamGraph): a streaming graph pipeline
    """
    if not transformations:
        raise GraphException('no transformations defined')

    graph = StreamGraph(storage)

    for operation in transformations:
        graph.add_node(operation)

    return graph


class StreamGraph(object):
    """ represent the streaming process graph """

    def __init__(self, storage):
        """ init graph """
        self.head = None
        self.storage = storage

    def __iter__(self):
        """ iterator for graph nodes

        Returns:
            self (StreamGraph): stream graph self
        """
        self.current = self.head
        return self

    def __next__(self):
        """ loop through nodes

        Raises:
            StopIteration: when all nodes are listed

        Returns:
            node (OperationNode): operation node
        """
        while self.current is not None:
            node = self.current
            self.current = self.current.next

            return node

        raise StopIteration

    def __repr__(self):
        """ print graph """
        nodes = [f'{node.operation.type}()' for node in self]
        return f'StreamGraph({"->".join(nodes)})'

    def add_node(self, operation):
        """ adds operation node to graph

        Args:
            operation (object): operation node
        """
        new_node = OperationNode(operation)

        # set storage backend
        new_node.operation.storage = self.storage

        if not self.head:
            self.head = new_node
            return

        last = None

        for node in self:
            last = node

        if new_node.operation.type == 'ReducerFunction' \
                and last.operation.type != 'KeyByFunction':
            raise GraphException(f'{new_node.operation.type} must be '
                                 'preceeded by a KeyByFunction node')

        last.next = new_node

    def run(self, event_collection):
        """ run graph by looping through each node and
            processing the event collection

        Args:
            event (event): event object
        """
        for node in self:
            event_collection = node.process(event_collection)

            if not event_collection:
                break

        return event_collection


class OperationNode(object):
    """ represent an operation as a node """

    def __init__(self, operation):
        """ init node """
        self.operation = operation
        self.next = None

    def process(self, event_collection):
        """ run operation over event """
        return self.operation(event_collection)



import pytest
import sqlite3

from fluxt.storage import LocalDatabase

DB_PATH = 'streaming_test.db'


@pytest.fixture
def local():
    return LocalDatabase(DB_PATH, 'my_stream_function')


def test_set_key(local):
    local.set_key('key', 1)

    con = sqlite3.connect(DB_PATH)
    cur = con.cursor()

    res = cur.execute('select * from keystore')
    rows = res.fetchall()

    cur.close()
    con.close()

    assert len(rows) == 1


def test_get_key(local):
    val = local.get_key('key')

    assert val == 1

    with pytest.raises(KeyError):
        local.get_key('test')


def test_reset(local):
    local.set_key('key', 1)

    local.reset()

    con = sqlite3.connect(DB_PATH)
    cur = con.cursor()

    res = cur.execute('select * from keystore')
    rows = res.fetchall()

    cur.close()
    con.close()

    assert len(rows) == 0



import pytest

from fluxt.app import Fluxt
from fluxt.operations import MapFunction
from fluxt.app.graph import GraphException


class Map(MapFunction):
    def map(self, event):
        return super().map(event)


@pytest.fixture()
def app():
    return Fluxt(name='Test Streaming App')


def test_repr(app):
    assert app.__repr__() == 'Fluxt(name="Test Streaming App")'


def test_stream(app):
    @app.stream()
    def stream(ds):
        ds.source_from_collection([1, 2, 3])
        ds.print()

    assert len(app.agents) == 1

    @app.stream()
    def stream_agent(ds):
        ds.source_from_collection([1, 2, 3])
        ds.print()

    assert len(app.agents) == 2


def test_run_no_transformations(app):
    @app.stream()
    def stream(ds):
        ds.source_from_collection([1, 2, 3])
        ds.print()

    with pytest.raises(GraphException) as e:
        app.run()

    assert 'no transformations defined' in str(e)


def test_run(app, capsys):
    @app.stream()
    def stream(ds):
        ds.source_from_collection([1, 2, 3])
        ds.print()

        ds.map(Map())

    app.run()
    out, err = capsys.readouterr()

    assert ['1', '2', '3'] == out.split()



from fluxt.app import Fluxt
import fluxt.operations as operations
from fluxt.sources import KafkaSource

# create a streaming app
fluxt = Fluxt(name='My Kafka Stream Processor')

@operations.filter()
def filter_strings(event):
    if 'badstring'.upper() in event:
        return False
    return True

@operations.key_by()
def key(event):
    return event, 1

@operations.flat_map()
def tokenize(event):
    return event.upper().split()

@operations.reducer()
def count(key, accum, event):
    if not accum:
        return event
    return accum + event

@fluxt.stream()
def stream_processor(data_stream):
    data_stream.add_source(KafkaSource('foobar', bootstrap_servers='localhost:9092'))

    data_stream.flat_map(tokenize) \
              .filter(filter_strings) \
              .key_by(key) \
              .reduce(count)

    data_stream.print()



from kafka import KafkaConsumer

from fluxt.sources.base import Source


class KafkaSource(Source):
    def __init__(self, topic, bootstrap_servers):
        self.bootstrap_servers = bootstrap_servers
        self.topic = topic

        self.consumer = KafkaConsumer(self.topic,
                                      bootstrap_servers=self.bootstrap_servers)

    def generate(self):
        for event in self.consumer:
            yield event.value.decode('utf-8')



class EventCollection(object):
    """ contain the event batch being processed """

    def __init__(self, base_event):
        """ init event collection """
        self.event_collection = [base_event]

    def __bool__(self):
        """ check if event collection is populated """
        if self.event_collection:
            return True
        return False

    @property
    def events(self):
        """ return list of events """
        return self.event_collection

    @events.setter
    def events(self, events):
        """ set list of events """
        self.event_collection = events



from fluxt.sources import CollectionSource


def test_collection_source():
    collection = ['1', '2', '3']
    source = CollectionSource(collection)
    assert source.collection == collection


def test_collection_source_generate():
    source = CollectionSource(['event'])

    for event in source.generate():
        assert event == 'event'



from fluxt import Fluxt

import fluxt.operations as operations

# create a streaming app
fluxt = Fluxt(name='My Stream Processor')

@operations.map()
def map_processor(event):
    return f'mapped-{event}'


@operations.filter()
def filter_processor(event):
    if 'event' in event:
        return True

    return False


@fluxt.stream()
def stream_processor(datastream):
    events = ['event1', 'event2', 'event3', 'bad']
    print(f'input events = {events}')

    datastream.source_from_collection(events)

    datastream.filter(filter_processor) \
          .map(map_processor)

    datastream.print()


import string

from fluxt.app import Fluxt

import fluxt.operations as operations

from fluxt.sources import FileSource
from fluxt.sinks import FileSink

# create a streaming app
fluxt = Fluxt(name='Word Count File')

@operations.flat_map()
def tokenizer(event):
    return event.lower().split()

@operations.map()
def remove_grammer(event):
    return event.translate(str.maketrans('', '', string.punctuation))

@operations.reducer()
def count_reducer(key, accum, event):
    if not accum:
        return 1
    return event + accum

@operations.key_by()
def key_index(event):
    return event, 1


@fluxt.stream()
def stream_processor(datastream):
    datastream.add_source(FileSource('examples/data/lorem_ipsum.txt'))

    datastream.pipeline(tokenizer, remove_grammer,
                            key_index, count_reducer)

    datastream.add_sink(FileSink('output.txt'))



from fluxt.storage.base import Base


class Memory(Base):
    def __init__(self):
        """ init memory storage """
        self.state = {}

    def reset(self):
        """ reset memory """
        self.state = {}

    def set_key(self, key, value):
        """ set key value """
        self.state[key] = value

    def get_key(self, key):
        """ get key value """
        return self.state[key]



from fluxt.sources.base import Source


class FileSource(Source):
    def __init__(self, file):
        """ init collection source """
        self.file = file

    def generate(self):
        """ generate events """
        with open(self.file, 'r', encoding='UTF-8') as f:
            for line in f.readlines():
                yield line.strip()



from fluxt.storage import Memory
from fluxt.app.datastream import DataStream


class Fluxt(object):
    """ Streaming App Container """

    def __init__(self, name=None, storage=None):
        """ init streaming app """
        self.name = name
        self.agents = []

        if not storage:
            storage = Memory()

        self.storage = storage

    def __repr__(self):
        """ streaming app repr """
        return f'Fluxt(name="{self.name}")'

    def stream(self, *args, **kwargs):
        """ add stream agent """
        def inner(func):
            self.agents.append(func)
            return func
        return inner

    def run(self):
        """ run streaming app """
        datastream = DataStream()

        datastream.storage = self.storage

        for agent in self.agents:
            agent(datastream)

        datastream.execute()



import pytest

from fluxt.operations import FlatMapFunction, flat_map
from fluxt.app.events import EventCollection


class FlatMap(FlatMapFunction):
    def flat_map(self, event):
        return event.split(' ')


class BadFlatMap(FlatMapFunction):
    pass


def test_flat_map_call():
    flat_map_function = FlatMap()
    event_collection = EventCollection(None)
    event_collection.events = ['text', 'text', 'some text']
    event_collection = flat_map_function(event_collection)

    assert len(event_collection.events) == 4
    assert event_collection.events == ['text', 'text', 'some', 'text']


def test_flat_map_type():
    flat_map_function = FlatMap()
    assert flat_map_function.type == 'FlatMapFunction'


def test_decorated_flat_map():
    @flat_map()
    def my_flat_map(event):
        return []

    event_collection = EventCollection(None)
    event_collection.events = [1, 2, 3]

    assert my_flat_map(event_collection).events == []


def test_flat_map_function():
    flat_map_function = FlatMap()
    event_collection = flat_map_function.flat_map('test text')

    assert len(event_collection) == 2
    assert event_collection == ['test', 'text']


def test_flat_map_function_abstract():
    with pytest.raises(TypeError):
        BadFlatMap()



import pytest

from fluxt.operations import FilterFunction, filter
from fluxt.app.events import EventCollection


class GoodFilterFunction(FilterFunction):
    def filter(self, event):
        return True


class BadFilterFunction(FilterFunction):
    pass


def test_filter_call():
    filter_function = GoodFilterFunction()
    event_collection = EventCollection(None)
    event_collection.events = [1, 2, 3]
    event_collection = filter_function(event_collection)

    assert event_collection.events == [1, 2, 3]


def test_filter_type():
    filter_function = GoodFilterFunction()
    assert filter_function.type == 'FilterFunction'


def test_decorated_filter():
    @filter()
    def my_filter(event):
        return True

    event_collection = EventCollection(None)
    event_collection.events = [1, 2, 3]

    assert my_filter(event_collection)


def test_filter_function():
    filter_function = GoodFilterFunction()
    assert filter_function.filter({'event': 'test'})


def test_filter_function_abstract():
    with pytest.raises(TypeError):
        BadFilterFunction()



from fluxt.operations.filter import FilterFunction, \
    filter_function_generator
from fluxt.operations.map import MapFunction, \
    map_function_generator
from fluxt.operations.flat_map import FlatMapFunction, \
    flat_map_function_generator
from fluxt.operations.reduce import ReducerFunction, \
    reducer_function_generator
from fluxt.operations.key import KeyByFunction, \
    key_by_function_generator


def filter(*args, **kwargs):
    """ filter function decorator """
    def inner(func):
        return filter_function_generator(func)
    return inner


def flat_map(*args, **kwargs):
    """ flat map function decorator """
    def inner(func):
        return flat_map_function_generator(func)
    return inner


def map(*args, **kwargs):
    """ map function decorator """
    def inner(func):
        return map_function_generator(func)
    return inner


def reducer(*args, **kwargs):
    """ reducer function decorator """
    def inner(func):
        return reducer_function_generator(func)
    return inner


def key_by(*args, **kwargs):
    """ key by function decorator """
    def inner(func):
        return key_by_function_generator(func)
    return inner



from abc import ABC, abstractmethod

from fluxt.operations.base import Operation


def flat_map_function_generator(func):
    class TempFlatMap(FlatMapFunction):
        def flat_map(self, event):
            return func(event)

    return TempFlatMap()


class FlatMapFunction(Operation):
    """ flat map function class """

    def process_batch(self, events):
        """ passes a batch of events to flat map

        Args:
            events (list): list of events

        Returns:
            batch (list): list of modified events
        """
        batch = []
        for event in events:
            batch += self.flat_map(event)

        return batch

    @property
    def type(self):
        """ return the operation type """
        return self.__class__.__base__.__name__

    @abstractmethod
    def flat_map(self, event):
        """ implements a flat map function

        Args:
            event (object): event object

        Returns:
            collection (list): list of events
        """
        return []



from fluxt.sinks import FileSink

from fluxt.app.events import EventCollection


def test_file_sink(monkeypatch):
    agg = []

    def mock_file_write(event):
        agg.append(event)

    sink = FileSink('output.txt')
    event_collection = EventCollection({'test': 'event'})
    monkeypatch.setattr(sink.file, 'write', mock_file_write)

    sink.pipe(event_collection)

    assert len(agg) == len(event_collection.events)



import pytest

from fluxt.storage import Memory


@pytest.fixture
def memory():
    return Memory()


def test_set_key(memory):
    memory.set_key('key', 1)

    assert memory.state['key'] == 1


def test_get_key(memory):
    memory.set_key('key', 1)

    val = memory.get_key('key')

    assert val == 1

    with pytest.raises(KeyError):
        memory.get_key('test')


def test_reset(memory):
    memory.set_key('key', 1)

    memory.reset()

    assert not memory.state



import pytest

from fluxt.operations import MapFunction, map
from fluxt.app.events import EventCollection


class GoodMapFunction(MapFunction):
    def map(self, event):
        return event


class KeyedMapFunction(MapFunction):
    def map(self, event):
        return self.keyed_event(event, 1)


class BadMapFunction(MapFunction):
    pass


def test_map_call():
    test_map = GoodMapFunction()
    event_collection = EventCollection(None)
    event_collection.events = [1, 2, 3]
    event_collection = test_map(event_collection)

    assert event_collection.events == [1, 2, 3]


def test_map_type():
    test_map = GoodMapFunction()
    assert test_map.type == 'MapFunction'


def test_decorated_map():
    @map()
    def my_map(event):
        return event

    event_collection = EventCollection(None)
    event_collection.events = [1, 2, 3]

    assert my_map(event_collection).events == [1, 2, 3]


def test_map_function():
    test_map = GoodMapFunction()
    event = test_map.map({'event': 'data'})

    assert event == {'event': 'data'}


def test_map_function_abstract():
    with pytest.raises(TypeError):
        BadMapFunction()



from abc import abstractmethod

from fluxt.operations.base import Operation


def reducer_function_generator(func):
    class TempReducer(ReducerFunction):
        def reduce(self, key, reduced, event):
            return func(key, reduced, event)

    return TempReducer()


class ReducerFunction(Operation):
    """ reducer function class """

    def process_batch(self, events):
        """ passes a batch of events to reduce

        Args:
            events (list): list of events

        Returns:
            batch (list): list of modified events
        """
        batch = []
        for event in events:
            key, value = event

            # attempt collect reduced
            try:
                reduced = self.storage_backend.get_key(key)
            except KeyError:
                reduced = None

            reduced = self.reduce(key, reduced, value)

            self.storage_backend.set_key(key, reduced)
            batch.append((key, reduced))

        return batch

    @property
    def type(self):
        """ return the operation type """
        return self.__class__.__base__.__name__

    @abstractmethod
    def reduce(self, key, reduced, event):
        """ abstract reducer """
        return key, event



import sys

from fluxt.sinks.base import Sink


class FileSink(Sink):
    def __init__(self, destination):
        """ init file sink """
        self.destination = destination
        self.file = open(self.destination, 'w')

    def pipe(self, event_collection):
        """ pipe event to file output """
        for event in event_collection.events:
            self.file.write(f'{event}\n')



import pytest

from fluxt.operations import ReducerFunction, reducer
from fluxt.storage import Memory
from fluxt.app.events import EventCollection


class GoodReduceFunction(ReducerFunction):
    def reduce(self, key, reduced, event):
        if reduced is None:
            return 1

        return reduced + event


class BadReduceFunction(ReducerFunction):
    pass


@pytest.fixture
def memory():
    return Memory()


def test_reduce_call(memory):
    test_reduce = GoodReduceFunction()
    test_reduce.storage = memory

    event_collection = EventCollection(None)
    event_collection.events = [('key', 1), ('key', 2), ('other', 1)]

    event_collection = test_reduce(event_collection)

    assert event_collection.events == [('key', 1), ('key', 3), ('other', 1)]


def test_reduce_type():
    test_reduce = GoodReduceFunction()
    assert test_reduce.type == 'ReducerFunction'


def test_decorated_map(memory):
    @reducer()
    def my_reducer(key, accum, event):
        if not accum:
            return 1
        return event + accum

    my_reducer.storage = memory

    event_collection = EventCollection(None)
    event_collection.events = [('key', 1), ('key', 2), ('other', 1)]

    assert my_reducer(event_collection).events == [('key', 1), ('key', 3),
                                                   ('other', 1)]


def test_reduce_function():
    test_reduce = GoodReduceFunction()
    event = test_reduce.reduce('key', 1, 2)

    assert event == 3


def test_reduce_function_abstract():
    with pytest.raises(TypeError):
        BadReduceFunction()



from abc import abstractmethod

from fluxt.operations.base import Operation


def key_by_function_generator(func):
    class TempKeyBy(KeyByFunction):
        def key_by(self, event):
            return func(event)

    return TempKeyBy()


class KeyByFunction(Operation):
    """ key the event stream """

    def process_batch(self, events):
        """ passes a batch of events to key

        Args:
            events (list): list of events

        Returns:
            batch (list): list of modified events
        """

        batch = []

        for event in events:
            key, value = self.key_by(event)
            batch.append((key, value))

        return batch

    @property
    def type(self):
        """ return the operation type """
        return self.__class__.__base__.__name__

    @abstractmethod
    def key_by(self, event):
        return True



from fluxt.sinks import StdoutSink

from fluxt.app.events import EventCollection


def test_stdout_sink(capsys):
    stdout = StdoutSink()
    event_collection = EventCollection({'test': 'event'})
    stdout.pipe(event_collection)
    out, err = capsys.readouterr()

    assert str(event_collection.events[0]) in out



from fluxt.sources.base import Source


class CollectionSource(Source):
    def __init__(self, collection):
        """ init collection source """
        self.collection = collection

    def generate(self):
        """ generate events """
        for event in self.collection:
            yield event



import pytest

from fluxt.app.graph import OperationNode, StreamGraph, \
    graph_generator, GraphException
from fluxt.operations import FilterFunction, MapFunction, \
    ReducerFunction, KeyByFunction
from fluxt.app.events import EventCollection
from fluxt.storage import Memory


class Map(MapFunction):
    def map(self, event):
        return super().map(event)


class Filter(FilterFunction):
    def filter(self, event):
        return super().filter(event)


class Reducer(ReducerFunction):
    def reduce(self, accumalator, event):
        pass


class KeyBy(KeyByFunction):
    def key_by(self, event):
        return super().key_by(event)


@pytest.fixture()
def storage():
    return Memory()


def test_stream_graph_generator(storage):
    graph = graph_generator([Filter(), Map()], storage)
    assert isinstance(graph, StreamGraph)


def test_stream_graph_repr(storage):
    graph = graph_generator([Filter(), Map()], storage)
    assert graph.__repr__() == 'StreamGraph(FilterFunction()->MapFunction())'


def test_stream_graph_iter(storage):
    graph = graph_generator([Filter(), Map()], storage)

    nodes = [node for node in graph]

    assert len(nodes) == 2
    assert nodes[0].operation.type == 'FilterFunction'
    assert nodes[1].operation.type == 'MapFunction'

    graph = graph_generator([Filter(), Map(), Map()], storage)
    nodes = [node for node in graph]

    assert len(nodes) == 3


def test_stream_graph_add_node(storage):
    graph = StreamGraph(storage)
    graph.add_node(Filter())

    assert graph.head
    assert graph.head.next is None

    graph.add_node(Map())

    assert graph.head.next

    with pytest.raises(GraphException):
        graph.add_node(Reducer())


def test_stream_graph_run(storage):
    event = {'test': 'event'}
    graph = graph_generator([Map(), Filter()], storage)

    event_collection = graph.run(EventCollection(event))

    assert len(event_collection.events) == 1
    assert event_collection.events[0] == event


def test_stream_graph_run_empty_events(storage):
    event = {'test': 'event'}

    class FilterAll(FilterFunction):
        def filter(self, event):
            return False

    graph = graph_generator([FilterAll()], storage)

    event_collection = graph.run(EventCollection(event))

    assert not event_collection


def test_operation_node_process():
    node = OperationNode(Map())
    event = {'test': 'event'}

    results = node.process(EventCollection(event))

    assert isinstance(results, EventCollection)
    assert results.event_collection == [event]



import logging

from fluxt.app.operations import DataStreamOperations
from fluxt.app.events import EventCollection
from fluxt.app.graph import graph_generator
from fluxt.sources import CollectionSource
from fluxt.sinks import StdoutSink

logger = logging.getLogger(__name__)


class DataStreamException(Exception):
    """ DataStream error """
    pass


class DataStream(DataStreamOperations):
    def __init__(self, source=None, sink=None, storage=None):
        """ DataStream class

        Args:
            source (Source, optional): datastream source. Defaults to None.
            sink (Sink, optional): datastream sink. Defaults to None.
        """
        self.transformations = []
        self.source = source
        self.sink = sink
        self.storage = storage

    def source_from_collection(self, collection):
        """ source from collection type """
        self.source = CollectionSource(collection)

    def print(self):
        """ sink to stdout """
        self.sink = StdoutSink()

    def add_source(self, source):
        """ add new source for ds

        Args:
            source (Source): datastream source
        """
        self.source = source

    def add_sink(self, sink):
        """ add new sink for ds

        Args:
            sink (Sink): datastream sink
        """
        self.sink = sink

    def execute(self):
        """ execute datastream transformations """
        if not self.source:
            raise DataStreamException('DataStream source not defined')

        if not self.sink:
            raise DataStreamException('DataStream sink not defined')

        execution_graph = graph_generator(self.transformations, self.storage)

        for event in self.source.generate():
            event_collection = execution_graph.run(EventCollection(event))
            self.sink.pipe(event_collection)



from fluxt.sources import FileSource


def test_file_source():
    file_path = 'examples/data/lorem_ipsum.txt'
    source = FileSource(file_path)
    assert source.file == file_path


def test_collection_source_generate():
    file_path = 'examples/data/lorem_ipsum.txt'
    source = FileSource(file_path)

    f = open(file_path)

    lines = f.readlines()
    events = []

    for event in source.generate():
        events.append(event)

    assert len(lines) == len(events)

    for i in range(len(lines)):
        assert lines[i].strip() == events[i]



from abc import abstractmethod

from fluxt.operations.base import Operation


def filter_function_generator(func):
    class TempFilter(FilterFunction):
        def filter(self, event):
            return func(event)

    return TempFilter()


class FilterFunction(Operation):
    """ filter function class """

    def process_batch(self, events):
        """ passes a batch of events to filter

        Args:
            events (list): list of events

        Returns:
            batch (list): list of modified events
        """
        batch = []
        for event in events:
            if self.filter(event):
                batch.append(event)

        return batch

    @property
    def type(self):
        """ return the operation type """
        return self.__class__.__base__.__name__

    @abstractmethod
    def filter(self, event):
        """ implements a filter function

        Args:
            event (object): event object

        Returns:
            boolean: bool if event should be returned
        """
        return True
